"""
===============================
Parallel Transport Tractography
===============================
Parallel Transport Tractography (PTT) [Aydogan2021]_

Let's start by importing the necessary modules.
"""

from dipy.io.streamline import save_trk
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.data import get_sphere
from dipy.direction import PTTDirectionGetter
from dipy.reconst.shm import CsaOdfModel
from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.reconst.csdeconv import (ConstrainedSphericalDeconvModel,
                                   auto_response_ssst)
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.streamline import Streamlines
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.viz import window, actor, colormap, has_fury


# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames('stanford_hardi')
label_fname = get_fnames('stanford_labels')

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs)

seed_mask = (labels == 2)
white_matter = (labels == 1) | (labels == 2)
seeds = utils.seeds_from_mask(seed_mask, affine, density=1)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order=6)
csd_fit = csd_model.fit(data, mask=white_matter)

###############################################################################
# We use the GFA of the CSA model to build a stopping criterion.

csa_model = CsaOdfModel(gtab, sh_order=6)
gfa = csa_model.fit(data, mask=white_matter).gfa
stopping_criterion = ThresholdStoppingCriterion(gfa, .25)

###############################################################################
# Prepare the PTT direction getter using the fiber ODF (FOD) obtain with CSD.
# Start the local tractography using PTT direction getter.

sphere = get_sphere(name='repulsion724')
fod = csd_fit.odf(sphere)
pmf = fod.clip(min=0)
ptt_dg = PTTDirectionGetter.from_pmf(pmf, max_angle=15, probe_length=0.5,
                                     sphere=sphere)

# Parallel Transport Tractography
streamline_generator = LocalTracking(direction_getter=ptt_dg,
                                     stopping_criterion=stopping_criterion,
                                     seeds=seeds,
                                     affine=affine,
                                     step_size=0.2)
streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_ptt_dg_pmf.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colormap.line_colors(streamlines)))
    window.record(scene, out_path='tractogram_ptt_dg_pmf.png',
                  size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using ptt direction getter from PMF
#
#
#
# References
# ----------
# .. [Aydogan2021] Aydogan DB, Shi Y. Parallel Transport Tractography. IEEE
#     Trans Med Imaging. 2021 Feb;40(2):635-647. doi: 10.1109/TMI.2020.3034038.
#     Epub 2021 Feb 2. PMID: 33104507; PMCID: PMC7931442.



"""Additional Command-line interface for spin."""
import os
import shutil

import click
from spin import util
from spin.cmds import meson


# From scipy: benchmarks/benchmarks/common.py
def _set_mem_rlimit(max_mem=None):
    """Set address space rlimit."""
    import resource
    import psutil

    mem = psutil.virtual_memory()

    if max_mem is None:
        max_mem = int(mem.total * 0.7)
    cur_limit = resource.getrlimit(resource.RLIMIT_AS)
    if cur_limit[0] > 0:
        max_mem = min(max_mem, cur_limit[0])

    try:
        resource.setrlimit(resource.RLIMIT_AS, (max_mem, cur_limit[1]))
    except ValueError:
        # on macOS may raise: current limit exceeds maximum limit
        pass


def _commit_to_sha(commit):
    p = util.run(['git', 'rev-parse', commit], output=False, echo=False)
    if p.returncode != 0:
        raise click.ClickException(
                f'Could not find SHA matching commit `{commit}`')

    return p.stdout.decode('ascii').strip()


def _dirty_git_working_dir():
    # Changes to the working directory
    p0 = util.run(['git', 'diff-files', '--quiet'])

    # Staged changes
    p1 = util.run(['git', 'diff-index', '--quiet', '--cached', 'HEAD'])

    return (p0.returncode != 0 or p1.returncode != 0)


def _run_asv(cmd):
    # Always use ccache, if installed
    PATH = os.environ['PATH']
    # EXTRA_PATH = os.pathsep.join([
    #     '/usr/lib/ccache', '/usr/lib/f90cache',
    #     '/usr/local/lib/ccache', '/usr/local/lib/f90cache'
    # ])
    env = os.environ
    env['PATH'] = f'EXTRA_PATH:{PATH}'

    # Control BLAS/LAPACK threads
    env['OPENBLAS_NUM_THREADS'] = '1'
    env['MKL_NUM_THREADS'] = '1'

    # Limit memory usage
    try:
        _set_mem_rlimit()
    except (ImportError, RuntimeError):
        pass

    util.run(cmd, cwd='benchmarks', env=env)


@click.command()
@click.option(
    '--tests', '-t',
    default=None, metavar='TESTS', multiple=True,
    help="Which tests to run"
)
@click.option(
    '--compare', '-c',
    is_flag=True,
    default=False,
    help="Compare benchmarks between the current branch and main "
         "(unless other branches specified). "
         "The benchmarks are each executed in a new isolated "
         "environment."
)
@click.option(
    '--verbose', '-v', is_flag=True, default=False
)
@click.option(
    '--quick', '-q', is_flag=True, default=False,
    help="Run each benchmark only once (timings won't be accurate)"
)
@click.argument(
    'commits', metavar='',
    required=False,
    nargs=-1
)
@click.pass_context
def bench(ctx, tests, compare, verbose, quick, commits):
    """🏋 Run benchmarks.

    \b
    Examples:

    \b
    $ spin bench -t bench_lib
    $ spin bench -t bench_random.Random
    $ spin bench -t Random -t Shuffle

    Two benchmark runs can be compared.
    By default, `HEAD` is compared to `main`.
    You can also specify the branches/commits to compare:

    \b
    $ spin bench --compare
    $ spin bench --compare main
    $ spin bench --compare main HEAD

    You can also choose which benchmarks to run in comparison mode:

    $ spin bench -t Random --compare
    """
    if not commits:
        commits = ('main', 'HEAD')
    elif len(commits) == 1:
        commits = commits + ('HEAD',)
    elif len(commits) > 2:
        raise click.ClickException(
            'Need a maximum of two revisions to compare'
        )

    bench_args = []
    for t in tests:
        bench_args += ['--bench', t]

    if verbose:
        bench_args = ['-v'] + bench_args

    if quick:
        bench_args = ['--quick'] + bench_args

    if not compare:
        # No comparison requested; we build and benchmark the current version

        click.secho(
            "Invoking `build` prior to running benchmarks:",
            bold=True, fg="bright_green"
        )
        ctx.invoke(meson.build)

        meson._set_pythonpath()

        p = util.run(
            ['python', '-c', 'import dipy; print(dipy.__version__)'],
            cwd='benchmarks',
            echo=False,
            output=False
        )
        os.chdir('..')

        dipy_ver = p.stdout.strip().decode('ascii')
        click.secho(
            f'Running benchmarks on DIPY {dipy_ver}',
            bold=True, fg="bright_green"
        )
        cmd = [
            'asv', 'run', '--dry-run', '--show-stderr', '--python=same'
        ] + bench_args
        _run_asv(cmd)
    else:
        # Ensure that we don't have uncommited changes
        commit_a, commit_b = [_commit_to_sha(c) for c in commits]

        if commit_b == 'HEAD' and _dirty_git_working_dir():
            click.secho(
                "WARNING: you have uncommitted changes --- "
                "these will NOT be benchmarked!",
                fg="red"
            )

        cmd_compare = [
            'asv', 'continuous', '--factor', '1.05',
        ] + bench_args + [commit_a, commit_b]
        _run_asv(cmd_compare)


@click.command()
def clean():
    """🧹 Remove build and install folder."""
    build_dir = "build"
    install_dir = "build-install"
    print(f"Removing `{build_dir}`")
    if os.path.isdir(build_dir):
            shutil.rmtree(build_dir)
    print(f"Removing `{install_dir}`")
    if os.path.isdir(install_dir):
        shutil.rmtree(install_dir)


import pytest
from dipy.utils.optpkg import optional_package
from dipy.testing.decorators import set_random_number_generator
tf, have_tf, _ = optional_package('tensorflow', min_version='2.0.0')
sklearn, have_sklearn, _ = optional_package('sklearn.model_selection')
if have_tf and have_sklearn:
    from dipy.nn.cnn_1d_denoising import Cnn1DDenoiser


@pytest.mark.skipif(not have_tf or not have_sklearn,
                    reason='Requires TensorFlow and scikit-learn')
@set_random_number_generator()
def test_default_Cnn1DDenoiser_sequential(rng=None):
    # Create dummy data
    normal_img = rng.random((10, 10, 10, 30))
    nos_img = normal_img + rng.normal(loc=0.0, scale=0.1,
                                            size=normal_img.shape)
    x = rng.random((10, 10, 10, 30))
    # Test 1D denoiser
    model = Cnn1DDenoiser(30)
    model.compile(optimizer='adam', loss='mean_squared_error',
                  metrics=['accuracy'])
    epochs = 1
    hist = model.fit(nos_img, normal_img, epochs=epochs)
    data = model.predict(x)
    model.evaluate(nos_img, normal_img, verbose=2)
    _ = hist.history['accuracy'][0]
    assert data.shape == x.shape


@pytest.mark.skipif(not have_tf or not have_sklearn,
                    reason='Requires TensorFlow and scikit-learn')
@set_random_number_generator()
def test_default_Cnn1DDenoiser_flow(pytestconfig, rng):
    # Create dummy data
    normal_img = rng.random((10, 10, 10, 30))
    nos_img = normal_img + rng.normal(loc=0.0, scale=0.1,
                                            size=normal_img.shape)
    x = rng.random((10, 10, 10, 30))
    # Test 1D denoiser with flow API
    model = Cnn1DDenoiser(30)
    if pytestconfig.getoption('verbose') > 0:
        model.summary()
    model.compile(optimizer='adam', loss='mean_squared_error',
                  metrics=['accuracy'])
    epochs = 1
    hist = model.fit(nos_img, normal_img, epochs=epochs)
    _ = model.predict(x)
    model.evaluate(nos_img, normal_img, verbose=2)
    accuracy = hist.history['accuracy'][0]
    assert accuracy > 0



import warnings

import numpy as np
from scipy import stats

from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import actor
    from fury.utils import apply_affine


class SlicesVisualizer:
    def __init__(self, interactor, scene, data, affine=None,
                 world_coords=False, percentiles=[2, 98], rgb=False):

        self._interactor = interactor
        self._scene = scene
        self._data = data
        self._affine = affine

        if not world_coords:
            self._affine = np.eye(4)

        self._slice_actors = [None] * 3

        self._data_ndim = data.ndim
        self._data_shape = data.shape
        self._rgb = False

        vol_data = self._data

        if ((self._data_ndim == 4 and rgb and self._data_shape[-1] == 3)
                or self._data_ndim == 3):
            self._rgb = True and not self._data_ndim == 3
            self._int_range = np.percentile(vol_data, percentiles)
            _evaluate_intensities_range(self._int_range)
        else:
            if self._data_ndim == 4 and rgb and self._data_shape[-1] != 3:
                warnings.warn('The rgb flag is enabled but the color '
                              + 'channel information is not provided')
            vol_data = self._volume_calculations(percentiles)

        self._vol_max = np.max(vol_data)
        self._vol_min = np.min(vol_data)

        self._resliced_vol = None

        print(f'Original shape: {self._data_shape}')
        self._create_and_resize_actors(vol_data, self._int_range)
        print(f'Resized to RAS shape: {self._data_shape}')

        self._sel_slices = np.rint(
            np.asarray(self._data_shape[:3]) / 2).astype(int)

        self._add_slice_actors_to_scene(self._sel_slices)

        self._picker_callback = None
        self._picked_voxel_actor = None

    def _volume_calculations(self, percentiles):
        for i in range(self._data.shape[-1]):
            vol_data = self._data[..., i]
            self._int_range = np.percentile(vol_data, percentiles)
            if np.sum(np.diff(self._int_range)) != 0:
                break
            else:
                if i < self._data_shape[-1] - 1:
                    warnings.warn(
                        f'Volume N°{i} does not have any contrast. '
                        'Please, check the value ranges of your data. '
                        'Moving to the next volume.')
                else:
                    _evaluate_intensities_range(self._int_range)
        return vol_data

    def _add_slice_actors_to_scene(self, visible_slices):
        self._slice_actors[0].display_extent(
            visible_slices[0], visible_slices[0], 0, self._data_shape[1] - 1,
            0, self._data_shape[2] - 1)

        self._slice_actors[1].display_extent(
            0, self._data_shape[0] - 1, visible_slices[1], visible_slices[1],
            0, self._data_shape[2] - 1)

        self._slice_actors[2].display_extent(
            0, self._data_shape[0] - 1, 0, self._data_shape[1] - 1,
            visible_slices[2], visible_slices[2])

        for act in self._slice_actors:
            self._scene.add(act)

    def _create_and_resize_actors(self, vol_data, value_range):
        self._slice_actors[0] = actor.slicer(
            vol_data, affine=self._affine, value_range=value_range,
            interpolation='nearest')

        self._resliced_vol = self._slice_actors[0].resliced_array()

        self._slice_actors[1] = self._slice_actors[0].copy()
        self._slice_actors[2] = self._slice_actors[0].copy()

        for slice_actor in self._slice_actors:
            slice_actor.AddObserver(
                'LeftButtonPressEvent', self._left_click_picker_callback, 1.)

        if self._data_ndim == 4 and not self._rgb:
            self._data_shape = (
                self._resliced_vol.shape + (self._data.shape[-1],))
        else:
            self._data_shape = self._resliced_vol.shape

    def _left_click_picker_callback(self, obj, event):
        # TODO: Find out why this is not triggered when opacity < 1
        event_pos = self._interactor.GetEventPosition()

        obj.picker.Pick(event_pos[0], event_pos[1], 0, self._scene)

        i, j, k = obj.picker.GetPointIJK()
        res = self._resliced_vol[i, j, k]

        try:
            message = f'{res:.2f}'
        except TypeError:
            message = f'{res[0]:.2f} {res[1]:.2f} {res[2]:.2f}'
        message = f'({i}, {j}, {k}) = {message}'
        self._picker_callback(message)
        # TODO: Fix this
        # self._replace_picked_voxel_actor(i, j, k)

    def _replace_picked_voxel_actor(self, x, y, z):
        if self._picked_voxel_actor:
            self._scene.rm(self._picked_voxel_actor)
        pnt = np.asarray([[x, y, z]])
        pnt = apply_affine(self._affine, pnt)
        self._picked_voxel_actor = actor.dot(
            pnt, colors=(.9, .4, .0), dot_size=10)
        self._scene.add(self._picked_voxel_actor)

    def change_volume(self, prev_idx, next_idx, intensities, visible_slices):
        vol_data = self._data[..., prev_idx]
        # NOTE: Supported only in latests versions of scipy
        # percs = stats.percentileofscore(np.ravel(vol_data), intensities)
        perc_0 = stats.percentileofscore(np.ravel(vol_data), intensities[0])
        perc_1 = stats.percentileofscore(np.ravel(vol_data), intensities[1])
        vol_data = self._data[..., next_idx]
        value_range = np.percentile(vol_data, [perc_0, perc_1])
        if np.sum(np.diff(self._int_range)) == 0:
            return False

        self._int_range = value_range

        self._vol_max = np.max(vol_data)
        self._vol_min = np.min(vol_data)

        for slice_actor in self._slice_actors:
            self._scene.rm(slice_actor)

        self._create_and_resize_actors(vol_data, self._int_range)

        self._add_slice_actors_to_scene(visible_slices)

        return True

    def register_picker_callback(self, callback):
        self._picker_callback = callback

    @property
    def data_shape(self):
        return self._data_shape

    @property
    def intensities_range(self):
        return self._int_range

    @property
    def selected_slices(self):
        return self._sel_slices

    @property
    def slice_actors(self):
        return self._slice_actors

    @property
    def volume_max(self):
        return self._vol_max

    @property
    def volume_min(self):
        return self._vol_min

    @property
    def rgb(self):
        return self._rgb


def _evaluate_intensities_range(intensities_range):
    if np.sum(np.diff(intensities_range)) == 0:
        raise ValueError(
            'Your data does not have any contrast. Please, check the '
            'value range of your data.')



"""
=====================================================================
Reconstruction of the diffusion signal with the WMTI model
=====================================================================

DKI can also be used to derive concrete biophysical parameters by applying
microstructural models to DT and KT estimated from DKI. For instance,
Fieremans et al. [Fierem2011]_ showed that DKI can be used to
estimate the contribution of hindered and restricted diffusion for well-aligned
fibers - a model that was later referred to as the white matter tract integrity
WMTI technique [Fierem2013]_. The two tensors of WMTI can be also
interpreted as the influences of intra- and extra-cellular compartments and can
be used to estimate the axonal volume fraction and diffusion extra-cellular
tortuosity. According to previous studies [Fierem2012]_ [Fierem2013]_,
these latter measures can be used to distinguish processes of axonal loss from
processes of myelin degeneration.

In this example, we show how to process a dMRI dataset using
the WMTI model.

First, we import all relevant modules:
"""

import numpy as np
import matplotlib.pyplot as plt
import dipy.reconst.dki as dki
import dipy.reconst.dki_micro as dki_micro
from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.segment.mask import median_otsu
from scipy.ndimage import gaussian_filter

###############################################################################
# As the standard DKI, WMTI requires multi-shell data, i.e. data acquired from
# more than one non-zero b-value. Here, we use a fetcher to download a
# multi-shell dataset which was kindly provided by Hansen and Jespersen
# (more details about the data are provided in their paper [Hansen2016]_).

fraw, fbval, fbvec, t1_fname = get_fnames('cfin_multib')

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs)

###############################################################################
# For comparison, this dataset is pre-processed using the same steps used in
# the example for reconstructing DKI (see
# :ref:`sphx_glr_examples_built_reconstruction_reconst_dki.py`).

# data masking
maskdata, mask = median_otsu(data, vol_idx=[0, 1], median_radius=4, numpass=2,
                             autocrop=False, dilate=1)

# Smoothing
fwhm = 1.25
gauss_std = fwhm / np.sqrt(8 * np.log(2))
data_smooth = np.zeros(data.shape)
for v in range(data.shape[-1]):
    data_smooth[..., v] = gaussian_filter(data[..., v], sigma=gauss_std)

###############################################################################
# The WMTI model can be defined in DIPY by instantiating the
# 'KurtosisMicrostructureModel' object in the following way:

dki_micro_model = dki_micro.KurtosisMicrostructureModel(gtab)

###############################################################################
# Before fitting this microstructural model, it is useful to indicate the
# regions in which this model provides meaningful information (i.e. voxels of
# well-aligned fibers). Following Fieremans et al. [Fieremans2011]_, a simple
# way to select this region is to generate a well-aligned fiber mask based on
# the values of diffusion sphericity, planarity and linearity. Here we will
# follow these selection criteria for a better comparison of our figures with
# the original article published by Fieremans et al. [Fieremans2011]_.
# Nevertheless, it is important to note that voxels with well-aligned fibers
# can be selected based on other approaches such as using predefined regions
# of interest.

# Diffusion Tensor is computed based on the standard DKI model
dkimodel = dki.DiffusionKurtosisModel(gtab)
dkifit = dkimodel.fit(data_smooth, mask=mask)

# Initialize well aligned mask with ones
well_aligned_mask = np.ones(data.shape[:-1], dtype='bool')

# Diffusion coefficient of linearity (cl) has to be larger than 0.4, thus
# we exclude voxels with cl < 0.4.
cl = dkifit.linearity.copy()
well_aligned_mask[cl < 0.4] = False

# Diffusion coefficient of planarity (cp) has to be lower than 0.2, thus
# we exclude voxels with cp > 0.2.
cp = dkifit.planarity.copy()
well_aligned_mask[cp > 0.2] = False

# Diffusion coefficient of sphericity (cs) has to be lower than 0.35, thus
# we exclude voxels with cs > 0.35.
cs = dkifit.sphericity.copy()
well_aligned_mask[cs > 0.35] = False

# Removing nan associated with background voxels
well_aligned_mask[np.isnan(cl)] = False
well_aligned_mask[np.isnan(cp)] = False
well_aligned_mask[np.isnan(cs)] = False

###############################################################################
# Analogous to DKI, the data fit can be done by calling the ``fit`` function of
# the model's object as follows:

dki_micro_fit = dki_micro_model.fit(data_smooth, mask=well_aligned_mask)

###############################################################################
# The KurtosisMicrostructureFit object created by this ``fit`` function can
# then be used to extract model parameters such as the axonal water fraction
# and diffusion hindered tortuosity:

AWF = dki_micro_fit.awf
TORT = dki_micro_fit.tortuosity

###############################################################################
# These parameters are plotted below on top of the mean kurtosis maps:

MK = dkifit.mk(0, 3)

axial_slice = 9

fig1, ax = plt.subplots(1, 2, figsize=(9, 4),
                        subplot_kw={'xticks': [], 'yticks': []})

AWF[AWF == 0] = np.nan
TORT[TORT == 0] = np.nan

ax[0].imshow(MK[:, :, axial_slice].T, cmap=plt.cm.gray,
             interpolation='nearest', origin='lower')
im0 = ax[0].imshow(AWF[:, :, axial_slice].T, cmap=plt.cm.Reds, alpha=0.9,
                   vmin=0.3, vmax=0.7, interpolation='nearest', origin='lower')
fig1.colorbar(im0, ax=ax.flat[0])

ax[1].imshow(MK[:, :, axial_slice].T, cmap=plt.cm.gray,
             interpolation='nearest', origin='lower')
im1 = ax[1].imshow(TORT[:, :, axial_slice].T, cmap=plt.cm.Blues, alpha=0.9,
                   vmin=2, vmax=6, interpolation='nearest', origin='lower')
fig1.colorbar(im1, ax=ax.flat[1])

fig1.savefig('Kurtosis_Microstructural_measures.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Axonal water fraction (left panel) and tortuosity (right panel) values
# of well-aligned fiber regions overlaid on a top of a mean kurtosis all-brain
# image.
#
#
# References
# ----------
#
# .. [Fierem2011] Fieremans E, Jensen JH, Helpern JA (2011). White matter
#                 characterization with diffusion kurtosis imaging. NeuroImage
#                 58: 177-188
# .. [Fierem2012] Fieremans E, Jensen JH, Helpern JA, Kim S, Grossman RI,
#                 Inglese M, Novikov DS. (2012). Diffusion distinguishes
#                 between axonal loss and demyelination in brain white matter.
#                 Proceedings of the 20th Annual Meeting of the International
#                 Society for Magnetic Resonance Medicine; Melbourne,
#                 Australia. May 5-11.
# .. [Fierem2013] Fieremans, E., Benitez, A., Jensen, J.H., Falangola, M.F.,
#                 Tabesh, A., Deardorff, R.L., Spampinato, M.V., Babb, J.S.,
#                 Novikov, D.S., Ferris, S.H., Helpern, J.A., 2013. Novel
#                 white matter tract integrity metrics sensitive to Alzheimer
#                 disease progression. AJNR Am. J. Neuroradiol. 34(11),
#                 2105-2112. doi: 10.3174/ajnr.A3553
# .. [Hansen2016] Hansen, B, Jespersen, SN (2016). Data for evaluation of fast
#                 kurtosis strategies, b-value optimization and exploration of
#                 diffusion MRI contrast. Scientific Data 3: 160072
#                 doi:10.1038/sdata.2016.72



# dipy documentation build configuration file, created by
# sphinx-quickstart on Thu Feb  4 15:23:20 2010.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import re
import sys
from datetime import datetime

# Doc generation depends on being able to import dipy
try:
    import dipy
except ImportError:
    raise RuntimeError('Cannot import dipy, please investigate')

from packaging.version import Version
import sphinx
if Version(sphinx.__version__) < Version('2'):
    raise RuntimeError('Need sphinx >= 2 for numpydoc to work correctly')

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath('sphinxext'))

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.intersphinx',
              'sphinx.ext.todo',
              'sphinx.ext.coverage',
              'sphinx.ext.mathjax',
              'sphinx.ext.ifconfig',
              'sphinx.ext.autosummary',
              'prepare_gallery',
              'math_dollar',  # has to go before numpydoc
              'sphinx_gallery.gen_gallery',
              # 'numpydoc',
              'github',
              'sphinx_design',
]

numpydoc_show_class_members = True
numpydoc_class_members_toctree = False

# ghissue config
github_project_url = "https://github.com/dipy/dipy"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'dipy'
copyright = f"Copyright 2008-{datetime.now().year}, DIPY developers. Created using Grg Sphinx Theme and PyData Sphinx Theme."

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = dipy.__version__
# The full version, including alpha/beta/rc tags.
release = version

# Include common links
# We don't use this any more because it causes conflicts with the gitwash docs
#rst_epilog = open('links_names.inc', 'rt').read()

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of documents that shouldn't be included in the build.
#unused_docs = []

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_patterns = ['_build', 'examples', 'examples_revamped']

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = "grg_sphinx_theme"

# The style sheet to use for HTML and HTML Help pages. A file of that name
# must exist either in Sphinx' static/ path, or in one of the custom paths
# given in html_static_path.
html_style = 'css/dipy.css'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
  "switcher": {"json_url": "https://docs.dipy.org/dev/_static/version_switcher.json",
               "version_match": "1.8.0", },
  "check_switcher": False,
  "show_version_warning_banner": True,
  "navbar_end": ["search-field.html", "version-switcher", "navbar-icon-links.html"],
  "secondary_sidebar_items": ["page-toc"],
  "show_toc_level": 1,
  "navbar_center": ["components/navbar-links.html"],
  "navbar_links": [
     {
        "name": "Docs",
        "children": [
          {
            "name": "Overview",
            "url": "index",
          },
          {
            "name": "Tutorials",
            "url": "examples_built/index",
          },
          {
            "name": "Recipes",
            "url": "recipes",
          },
          {
            "name": "CLI / Workflows",
            "url": "interfaces/index",
          },
          {
            "name": "API",
            "url": "reference/index",
          },
          {
            "name": "CLI API",
            "url": "reference_cmd/index",
          }
        ]
     },
     {
        "name": "Workshops",
        "sections": [
          {
            "name": "Latest",
            "children": [
              {
                "name": "DIPY Workshop 2024",
                "url": "https://dipy.org/workshops/dipy-workshop-2024",
                "link_type": "external"
              }
            ]
          },
          {
            "name": "Past",
            "children": [
              {
                "name": "DIPY Workshop 2023",
                "url": "https://dipy.org/workshops/dipy-workshop-2023",
                "link_type": "external"
              },
              {
                "name": "DIPY Workshop 2022",
                "url": "https://dipy.org/workshops/dipy-workshop-2022",
                "link_type": "external"
              },
              {
                "name": "DIPY Workshop 2021",
                "url": "https://dipy.org/workshops/dipy-workshop-2021",
                "link_type": "external"
              },
              {
                "name": "DIPY Workshop 2020",
                "url": "https://dipy.org/workshops/dipy-workshop-2020",
                "link_type": "external"
              },
              {
                "name": "DIPY Workshop 2019",
                "url": "https://dipy.org/workshops/dipy-workshop-2019",
                "link_type": "external"
              },
            ]
          }
        ],
     },
     {
        "name": "Community",
        "sections": [
            {
              "name": "News",
              "children": [
                  {
                    "name": "Calendar",
                    "url": "https://dipy.org/calendar",
                    "link_type": "inter"
                  },
                  {
                    "name": "Newsletters",
                    "url": "https://mail.python.org/mailman3/lists/dipy.python.org/",
                    "link_type": "external"
                  },
                  {
                    "name": "Blog",
                    "url": "https://dipy.org/blog",
                    "link_type": "inter"
                  },
                  {
                    "name": "Youtube",
                    "url": "https://www.youtube.com/c/diffusionimaginginpython",
                    "link_type": "external"
                  }
              ]
            },
            {
              "name": "Help",
              "children": [
                  {
                    "name": "Live Chat (Gitter)",
                    "url": "https://app.gitter.im/#/room/%23dipy_dipy:gitter.im",
                    "link_type": "external"
                  },
                  {
                    "name": "Github Discussions",
                    "url": "https://github.com/dipy/dipy/discussions",
                    "link_type": "external"
                  }
              ]
            }
          ]
     },
     {
        "name": "About",
        "children": [
          {
            "name": "Team",
            "url": "https://dipy.org/team",
            "link_type": "inter"
          },
          {
            "name": "FAQ",
            "url": "faq",
          },
          {
            "name": "Mission Statement",
            "url": "user_guide/mission",
          },
          {
            "name": "Releases",
            "url": "stateoftheart",
          },
          {
            "name": "Cite",
            "url": "cite",
          },
          {
            "name": "Glossary",
            "url": "glossary",
          },
        ]
     },
  ],
  # To remove search icon
  "navbar_persistent": "",
  "icon_links": [
    {
      "name": "GitHub",
      "url": "https://github.com/dipy",
      "icon": "fa-brands fa-github"
    },
    {
      "name": "Twitter/X",
      "url": "https://twitter.com/dipymri",
      "icon": "fa-brands fa-twitter"
    },
    {
      "name": "YouTube",
      "url": "https://www.youtube.com/c/diffusionimaginginpython",
      "icon": "fa-brands fa-youtube"
    },
    {
      "name": "LinkedIn",
      "url": "https://www.linkedin.com/company/dipy/",
      "icon": "fa-brands fa-linkedin"
    },
  ],
  "logo": {
    "image_dark": "_static/images/logos/dipy-logo.png",
    "alt_text": "DIPY",
  },
  "footer_start": ["components/footer-sign-up.html"],
  "footer_signup_data": {
    "heading": "Never miss an update from us!",
    "sub_heading": "Don't worry! we are not going to spam you."
  },
  "footer_end": ["components/footer-sections.html"],
  "footer_links": [
    {
      "title": "About",
      "links": [
        {
          "name": "Developers",
          "link": "https://dipy.org/team",
          "link_type": "inter"
        },
        {
          "name": "Support",
          "link": "https://github.com/dipy/dipy/discussions",
          "link_type": "external"
        },
        {
          "name": "Download",
          "link": "user_guide/installation"
        },
        {
          "name": "Get Started",
          "link": "user_guide/getting_started"
        },
        {
          "name": "Tutorials",
          "link": "examples_built/index"
        },
        {
          "name": "Videos",
          "link": "https://www.youtube.com/c/diffusionimaginginpython",
          "link_type": "external"
        },
      ]
    }, {
      "title": "Friends",
      "links": [
        {
          "name": "Nipy Projects",
          "link": "http://nipy.org/",
          "link_type": "external"
        },
        {
          "name": "FURY",
          "link": "http://fury.gl/",
          "link_type": "external"
        },
        {
          "name": "Nibabel",
          "link": "http://nipy.org/nibabel",
          "link_type": "external"
        },
        {
          "name": "Tortoise",
          "link": "https://tortoise.nibib.nih.gov/",
          "link_type": "external"
        },
      ]
    }, {
      "title": "Support",
      "links": [
        {
          "name": "The department of Intelligent Systems Engineering of Indiana University",
          "link": "https://engineering.indiana.edu/",
          "link_type": "external"
        },
        {
          "name": "The National Institute of Biomedical Imaging and Bioengineering, NIH",
          "link": "https://www.nibib.nih.gov/",
          "link_type": "external"
        },
        {
          "name": "The Gordon and Betty Moore Foundation and the Alfred P. Sloan Foundation, through the University of Washington eScience Institute Data Science Environment",
          "link": "https://escience.washington.edu",
          "link_type": "external"
        },
        {
          "name": "Google supported DIPY through the Google Summer of Code Program during Summer 2015, 2016, 2018",
          "link": "https://summerofcode.withgoogle.com/",
          "link_type": "external"
        },
      ]
    }
  ],
  "footer_copyright": copyright,
}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/images/logos/dipy-logo.png"
html_baseurl = "https://dipy.org"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/logos/dipy-favicon.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'index': []}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
# Setting to false fixes double module listing under header
html_use_modindex = False

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# If nonempty, this is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = ''

# Output file base name for HTML help builder.
htmlhelp_basename = 'dipydoc'


# -- Options for LaTeX output --------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'dipy.tex', 'dipy Documentation',
   'Eleftherios Garyfallidis, Ian Nimmo-Smith, Matthew Brett', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# Additional stuff for the LaTeX preamble.
latex_preamble = r"""
\usepackage{amsfonts}
"""

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_use_modindex = True


# -- Options for sphinx gallery -------------------------------------------
from docimage_scrap import ImageFileScraper
from sphinx_gallery.sorting import ExplicitOrder
from prepare_gallery import folder_explicit_order

sc = ImageFileScraper()
ordered_folders = [f'examples_revamped/{f}' for f in folder_explicit_order()]

sphinx_gallery_conf = {
     'doc_module': ('dipy',),
     # path to your examples scripts
     'examples_dirs': ['examples_revamped', ],
     # path where to save gallery generated examples
     'gallery_dirs': ['examples_built', ],
     'subsection_order': ExplicitOrder(ordered_folders),
     'image_scrapers': (sc),
     'backreferences_dir': 'examples_built',
     'reference_url': {'dipy': None, },
     'abort_on_example_error': False,
     'filename_pattern': re.escape(os.sep),
     'default_thumb_file': "_static/images/logos/dipy_full_logo.png",
    #  'pypandoc': {'extra_args': ['--mathjax',]},
}

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3/', None)}




import numpy as np

from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab import (HorizonTab, build_label, color_double_slider,
                                  color_single_slider)

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import ui


class PeaksTab(HorizonTab):
    def __init__(self, peak_actor):
        self.__actor = peak_actor
        self.__name = 'Peaks'

        self.__tab_id = 0
        self.__tab_ui = None

        self.__toggler_label_view_mode = build_label(text='View Mode')

        """
        self.__interaction_mode_label.actor.AddObserver(
            'LeftButtonPressEvent', self.__change_interaction_mode_callback,
            1.)
        """

        self.__view_modes = {'Cross section': 0, 'Range': 1}
        self.__view_mode_toggler = ui.RadioButton(
            list(self.__view_modes), ['Cross section'], padding=1.5,
            font_size=16, font_family='Arial')

        self.__view_mode_toggler.on_change = self.__toggle_view_mode

        self.__slider_label_x = build_label(text='X Slice')
        self.__slider_label_y = build_label(text='Y Slice')
        self.__slider_label_z = build_label(text='Z Slice')

        # Initializing sliders
        min_centers = self.__actor.min_centers
        max_centers = self.__actor.max_centers

        length = 450
        lw = 3
        radius = 8
        fs = 16
        tt = '{value:.0f}'

        # Initializing cross section sliders
        cross_section = self.__actor.cross_section

        self.__slider_slice_x = ui.LineSlider2D(
            initial_value=cross_section[0], min_value=min_centers[0],
            max_value=max_centers[0], length=length, line_width=lw,
            outer_radius=radius, font_size=fs, text_template=tt)

        self.__slider_slice_y = ui.LineSlider2D(
            initial_value=cross_section[1], min_value=min_centers[1],
            max_value=max_centers[1], length=length, line_width=lw,
            outer_radius=radius, font_size=fs, text_template=tt)

        self.__slider_slice_z = ui.LineSlider2D(
            initial_value=cross_section[2], min_value=min_centers[2],
            max_value=max_centers[2], length=length, line_width=lw,
            outer_radius=radius, font_size=fs, text_template=tt)

        color_single_slider(self.__slider_slice_x)
        color_single_slider(self.__slider_slice_y)
        color_single_slider(self.__slider_slice_z)

        self.__slider_slice_x.on_change = self.__change_slice_x
        self.__slider_slice_y.on_change = self.__change_slice_y
        self.__slider_slice_z.on_change = self.__change_slice_z

        # Initializing cross section sliders
        low_ranges = self.__actor.low_ranges
        high_ranges = self.__actor.high_ranges

        self.__slider_range_x = ui.LineDoubleSlider2D(
            line_width=lw, outer_radius=radius, length=length,
            initial_values=(low_ranges[0], high_ranges[0]),
            min_value=min_centers[0], max_value=max_centers[0], font_size=fs,
            text_template=tt)

        self.__slider_range_y = ui.LineDoubleSlider2D(
            line_width=lw, outer_radius=radius, length=length,
            initial_values=(low_ranges[1], high_ranges[1]),
            min_value=min_centers[1], max_value=max_centers[1], font_size=fs,
            text_template=tt)

        self.__slider_range_z = ui.LineDoubleSlider2D(
            line_width=lw, outer_radius=radius, length=length,
            initial_values=(low_ranges[2], high_ranges[2]),
            min_value=min_centers[2], max_value=max_centers[2], font_size=fs,
            text_template=tt)

        color_double_slider(self.__slider_range_x)
        color_double_slider(self.__slider_range_y)
        color_double_slider(self.__slider_range_z)

        self.__slider_range_x.on_change = self.__change_range_x
        self.__slider_range_y.on_change = self.__change_range_y
        self.__slider_range_z.on_change = self.__change_range_z

    def __add_cross_section_sliders(self, x_pos=.12):
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_slice_x, (x_pos, .62))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_slice_y, (x_pos, .38))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_slice_z, (x_pos, .15))

    def __add_range_sliders(self, x_pos=.12):
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_range_x, (x_pos, .62))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_range_y, (x_pos, .38))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_range_z, (x_pos, .15))

    def __change_range_x(self, slider):
        val1 = slider.left_disk_value
        val2 = slider.right_disk_value
        lr = self.__actor.low_ranges
        hr = self.__actor.high_ranges
        self.__actor.display_extent(val1, val2, lr[1], hr[1], lr[2], hr[2])

    def __change_range_y(self, slider):
        val1 = slider.left_disk_value
        val2 = slider.right_disk_value
        lr = self.__actor.low_ranges
        hr = self.__actor.high_ranges
        self.__actor.display_extent(lr[0], hr[0], val1, val2, lr[2], hr[2])

    def __change_range_z(self, slider):
        val1 = slider.left_disk_value
        val2 = slider.right_disk_value
        lr = self.__actor.low_ranges
        hr = self.__actor.high_ranges
        self.__actor.display_extent(lr[0], hr[0], lr[1], hr[1], val1, val2)

    def __change_slice_x(self, slider):
        value = int(np.rint(slider.value))
        cs = self.__actor.cross_section
        self.__actor.display_cross_section(value, cs[1], cs[2])

    def __change_slice_y(self, slider):
        value = int(np.rint(slider.value))
        cs = self.__actor.cross_section
        self.__actor.display_cross_section(cs[0], value, cs[2])

    def __change_slice_z(self, slider):
        value = int(np.rint(slider.value))
        cs = self.__actor.cross_section
        self.__actor.display_cross_section(cs[0], cs[1], value)

    def __hide_cross_section_sliders(self):
        self.__slider_slice_x.set_visibility(False)
        self.__slider_slice_y.set_visibility(False)
        self.__slider_slice_z.set_visibility(False)

    def __hide_range_sliders(self):
        self.__slider_range_x.set_visibility(False)
        self.__slider_range_y.set_visibility(False)
        self.__slider_range_z.set_visibility(False)

    def __show_cross_section_sliders(self):
        self.__slider_slice_x.set_visibility(True)
        self.__slider_slice_y.set_visibility(True)
        self.__slider_slice_z.set_visibility(True)

    def __show_range_sliders(self):
        self.__slider_range_x.set_visibility(True)
        self.__slider_range_y.set_visibility(True)
        self.__slider_range_z.set_visibility(True)

    def __toggle_view_mode(self, radio):
        view_mode = self.__view_modes[radio.checked_labels[0]]
        # view_mode==0: Cross section - view_mode==1: Range
        if view_mode:
            self.__actor.display_extent(
                self.__actor.low_ranges[0], self.__actor.high_ranges[0],
                self.__actor.low_ranges[1], self.__actor.high_ranges[1],
                self.__actor.low_ranges[2], self.__actor.high_ranges[2])
            self.__hide_cross_section_sliders()
            self.__show_range_sliders()
        else:
            self.__actor.display_cross_section(
                self.__actor.cross_section[0], self.__actor.cross_section[1],
                self.__actor.cross_section[2])
            self.__hide_range_sliders()
            self.__show_cross_section_sliders()

    def build(self, tab_id, tab_ui):
        self.__tab_id = tab_id
        self.__tab_ui = tab_ui

        x_pos = .02

        self.__tab_ui.add_element(
            self.__tab_id, self.__toggler_label_view_mode, (x_pos, .85))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_x, (x_pos, .62))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_y, (x_pos, .38))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_z, (x_pos, .15))

        x_pos=.12

        self.__tab_ui.add_element(
            self.__tab_id, self.__view_mode_toggler, (x_pos, .80))

        # Default view of Peak Actor is range
        self.__add_range_sliders()
        self.__hide_range_sliders()

        # Changing initial view to cross section
        cross_section = self.__actor.cross_section
        self.__actor.display_cross_section(
            cross_section[0], cross_section[1], cross_section[2])
        self.__add_cross_section_sliders()

    @property
    def name(self):
        return self.__name



import os
from os.path import join as pjoin
from tempfile import TemporaryDirectory

import numpy as np
import numpy.testing as npt

from dipy.direction.peaks import PeaksAndMetrics
from dipy.data import default_sphere
from dipy.io.peaks import load_peaks, save_peaks, peaks_to_niftis
from dipy.testing.decorators import set_random_number_generator


@set_random_number_generator()
def test_io_peaks(rng):
    with TemporaryDirectory() as tmpdir:
        fname = 'test.pam5'

        pam = PeaksAndMetrics()
        pam.affine = np.eye(4)
        pam.peak_dirs = rng.random((10, 10, 10, 5, 3))
        pam.peak_values = np.zeros((10, 10, 10, 5))
        pam.peak_indices = np.zeros((10, 10, 10, 5))
        pam.shm_coeff = np.zeros((10, 10, 10, 45))
        pam.sphere = default_sphere
        pam.B = np.zeros((45, default_sphere.vertices.shape[0]))
        pam.total_weight = 0.5
        pam.ang_thr = 60
        pam.gfa = np.zeros((10, 10, 10))
        pam.qa = np.zeros((10, 10, 10, 5))
        pam.odf = np.zeros((10, 10, 10, default_sphere.vertices.shape[0]))

        save_peaks(pjoin(tmpdir, fname), pam)
        pam2 = load_peaks(pjoin(tmpdir, fname), verbose=True)
        npt.assert_array_equal(pam.peak_dirs, pam2.peak_dirs)

        pam2.affine = None

        fname2 = 'test2.pam5'
        save_peaks(pjoin(tmpdir, fname2), pam2, np.eye(4))
        pam2_res = load_peaks(pjoin(tmpdir, fname2), verbose=True)
        npt.assert_array_equal(pam.peak_dirs, pam2_res.peak_dirs)

        pam3 = load_peaks(pjoin(tmpdir, fname2), verbose=False)

        for attr in ['peak_dirs', 'peak_values', 'peak_indices',
                     'gfa', 'qa', 'shm_coeff', 'B', 'odf']:
            npt.assert_array_equal(getattr(pam3, attr),
                                   getattr(pam, attr))

        npt.assert_equal(pam3.total_weight, pam.total_weight)
        npt.assert_equal(pam3.ang_thr, pam.ang_thr)
        npt.assert_array_almost_equal(pam3.sphere.vertices,
                                      pam.sphere.vertices)

        fname3 = 'test3.pam5'
        pam4 = PeaksAndMetrics()
        npt.assert_raises((ValueError, AttributeError), save_peaks,
                          pjoin(tmpdir, fname3), pam4)

        fname4 = 'test4.pam5'
        del pam.affine
        save_peaks(pjoin(tmpdir, fname4), pam, affine=None)

        fname5 = 'test5.pkm'
        npt.assert_raises(OSError, save_peaks, pjoin(tmpdir, fname5), pam)

        pam.affine = np.eye(4)
        fname6 = 'test6.pam5'
        save_peaks(pjoin(tmpdir, fname6), pam, verbose=True)

        del pam.shm_coeff
        save_peaks(pjoin(tmpdir, fname6), pam, verbose=False)

        pam.shm_coeff = np.zeros((10, 10, 10, 45))
        del pam.odf
        save_peaks(pjoin(tmpdir, fname6), pam)
        pam_tmp = load_peaks(pjoin(tmpdir, fname6), True)
        npt.assert_equal(pam_tmp.odf, None)

        fname7 = 'test7.paw'
        npt.assert_raises(OSError, load_peaks, pjoin(tmpdir, fname7))

        del pam.shm_coeff
        save_peaks(pjoin(tmpdir, fname6), pam, verbose=True)

        fname_shm = 'shm.nii.gz'
        fname_dirs = 'dirs.nii.gz'
        fname_values = 'values.nii.gz'
        fname_indices = 'indices.nii.gz'
        fname_gfa = 'gfa.nii.gz'

        pam.shm_coeff = np.ones((10, 10, 10, 45))
        peaks_to_niftis(pam, fname_shm, fname_dirs, fname_values,
                        fname_indices, fname_gfa, reshape_dirs=False)

        os.path.isfile(fname_shm)
        os.path.isfile(fname_dirs)
        os.path.isfile(fname_values)
        os.path.isfile(fname_indices)
        os.path.isfile(fname_gfa)


@set_random_number_generator()
def test_io_save_peaks_error(rng):
    with TemporaryDirectory() as tmpdir:
        fname = 'test.pam5'

        pam = PeaksAndMetrics()

        npt.assert_raises(OSError, save_peaks, pjoin(tmpdir, 'test.pam'), pam)
        npt.assert_raises((ValueError, AttributeError), save_peaks,
                          pjoin(tmpdir, fname), pam)

        pam.affine = np.eye(4)
        pam.peak_dirs = rng.random((10, 10, 10, 5, 3))
        pam.peak_values = np.zeros((10, 10, 10, 5))
        pam.peak_indices = np.zeros((10, 10, 10, 5))
        pam.shm_coeff = np.zeros((10, 10, 10, 45))
        pam.sphere = default_sphere
        pam.B = np.zeros((45, default_sphere.vertices.shape[0]))
        pam.total_weight = 0.5
        pam.ang_thr = 60
        pam.gfa = np.zeros((10, 10, 10))
        pam.qa = np.zeros((10, 10, 10, 5))
        pam.odf = np.zeros((10, 10, 10, default_sphere.vertices.shape[0]))



"""
====================================================
Extracting AFQ tract profiles from segmented bundles
====================================================

In this example, we will extract the values of a statistic from a
volume, using the coordinates along the length of a bundle. These are called
`tract profiles`

One of the challenges of extracting tract profiles is that some of the
streamlines in a bundle may diverge significantly from the bundle in some
locations. To overcome this challenge, we will use a strategy similar to that
described in [Yeatman2012]_: We will weight the contribution of each streamline
to the bundle profile based on how far this streamline is from the mean
trajectory of the bundle at that location.

"""

import dipy.stats.analysis as dsa
import dipy.tracking.streamline as dts
from dipy.segment.clustering import QuickBundles
from dipy.segment.metricspeed import AveragePointwiseEuclideanMetric
from dipy.segment.featurespeed import ResampleFeature
from dipy.data.fetcher import get_two_hcp842_bundles
import dipy.data as dpd
from dipy.io.streamline import load_trk
from dipy.io.image import load_nifti
import matplotlib.pyplot as plt
import numpy as np
import os.path as op

###############################################################################
# To get started, we will grab the bundles.

bundles_path = dpd.fetch_bundles_2_subjects()
bundles_folder = bundles_path[1]

cst_l_file = op.join(bundles_folder, "bundles_2_subjects", "subj_2", "bundles",
                     "bundles_cst.left.trk")
af_l_file = op.join(bundles_folder, "bundles_2_subjects", "subj_2", "bundles",
                    "bundles_af.left.trk")


###############################################################################
# Either way, we can use the `dipy.io` API to read in the bundles from file.
# `load_trk` returns both the streamlines, as well as header information.

cst_l = load_trk(cst_l_file, "same", bbox_valid_check=False).streamlines
af_l = load_trk(af_l_file, "same", bbox_valid_check=False).streamlines

###############################################################################
# In the next step, we need to make sure that all the streamlines in each
# bundle are oriented the same way. For example, for the CST, we want to make
# sure that all the bundles have their cortical termination at one end of the
# streamline.
# This is that when we later extract values from a volume, we won't have
# different streamlines going in opposite directions.
#
# To orient all the streamlines in each bundles, we will create standard
# streamlines, by finding the centroids of the left AF and CST bundle models.
#
# The advantage of using the model bundles is that we can use the same
# standard for different subjects, which means that we'll get roughly the
# same orientation

model_af_l_file, model_cst_l_file = get_two_hcp842_bundles()

model_af_l = load_trk(model_af_l_file, "same",
                      bbox_valid_check=False).streamlines
model_cst_l = load_trk(model_cst_l_file, "same",
                       bbox_valid_check=False).streamlines


feature = ResampleFeature(nb_points=100)
metric = AveragePointwiseEuclideanMetric(feature)

###############################################################################
# Since we are going to include all of the streamlines in the single cluster
# from the streamlines, we set the threshold to `np.inf`. We pull out the
# centroid as the standard.

qb = QuickBundles(np.inf, metric=metric)

cluster_cst_l = qb.cluster(model_cst_l)
standard_cst_l = cluster_cst_l.centroids[0]

cluster_af_l = qb.cluster(model_af_l)
standard_af_l = cluster_af_l.centroids[0]

###############################################################################
# We use the centroid streamline for each atlas bundle as the standard to
# orient all of the streamlines in each bundle from the individual subject.
# Here, the affine used is the one from the transform between the atlas and
# individual tractogram. This is so that the orienting is done relative to the
# space of the individual, and not relative to the atlas space.

oriented_cst_l = dts.orient_by_streamline(cst_l, standard_cst_l)
oriented_af_l = dts.orient_by_streamline(af_l, standard_af_l)

###############################################################################
# Read volumetric data from an image corresponding to this subject.
#
# For the purpose of this, we've extracted only the FA within the bundles in
# question, but in real use, this is where you would add the FA map of your
# subject.

files, folder = dpd.fetch_bundle_fa_hcp()

fa, fa_affine = load_nifti(op.join(folder, "hcp_bundle_fa.nii.gz"))

###############################################################################
# Calculate weights for each bundle:

w_cst_l = dsa.gaussian_weights(oriented_cst_l)
w_af_l = dsa.gaussian_weights(oriented_af_l)

###############################################################################
# And then use the weights to calculate the tract profiles for each bundle

profile_cst_l = dsa.afq_profile(fa, oriented_cst_l, fa_affine,
                                weights=w_cst_l)

profile_af_l = dsa.afq_profile(fa, oriented_af_l, fa_affine,
                               weights=w_af_l)

fig, (ax1, ax2) = plt.subplots(1, 2)

ax1.plot(profile_cst_l)
ax1.set_ylabel("Fractional anisotropy")
ax1.set_xlabel("Node along CST")
ax2.plot(profile_af_l)
ax2.set_xlabel("Node along AF")
fig.savefig("tract_profiles")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Bundle profiles for the fractional anisotropy in the left CST (left) and
# left AF (right).
#
#
#
# References
# ----------
#
# .. [Yeatman2012] Yeatman, Jason D., Robert F. Dougherty, Nathaniel J. Myall,
#     Brian A. Wandell, and Heidi M. Feldman. 2012. "Tract Profiles of White
#     Matter Properties: Automating Fiber-Tract Quantification" PloS One 7
#     (11): e49790.
#
# .. [Garyfallidis17] Garyfallidis et al. Recognition of white matter bundles
#    using local and global streamline-based registration and clustering,
#    Neuroimage, 2017.
#
# .. [Garyfallidis12] Garyfallidis E. et al., QuickBundles a method for
#    tractography simplification, Frontiers in Neuroscience, vol 6, no 175,
#    2012.



def check_img_shapes(images):
    """
    Checks if the images have same shapes
    """

    # No need to check if there is less than 2 images
    if len(images) < 2:
        return True
    base_shape = images[0][0].shape[:3]
    for img in images:
        data, _ = img
        if base_shape != data.shape[:3]:
            return False
    return True



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False

cimport numpy as cnp

from libc.stdlib cimport rand, srand, RAND_MAX
from libc.math cimport sqrt


# Replaces a numpy.searchsorted(arr, number, 'right')
cdef int where_to_insert(
        cnp.float_t* arr,
        cnp.float_t number,
        int size) nogil

cdef void cumsum(
        cnp.float_t* arr_in,
        cnp.float_t* arr_out,
        int N) nogil

cdef void copy_point(
        double * a,
        double * b) nogil

cdef void scalar_muliplication_point(
        double * a,
        double scalar) nogil

cdef double norm(
        double * v) nogil

cdef double dot(
        double * v1,
        double * v2) nogil

cdef void normalize(
        double * v) nogil

cdef void cross(
        double * out,
        double * v1,
        double * v2) nogil

cdef void random_vector(
        double * out) nogil

cdef void random_perpendicular_vector(
        double * out,
        double * v) nogil

cpdef (double, double) random_point_within_circle(
        double r) nogil

cpdef double random() nogil

cpdef void seed(cnp.npy_uint32 s) nogil




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Decorators for dipy tests
"""

import re
import os
import platform
import inspect
import numpy as np

SKIP_RE = re.compile(r"(\s*>>>.*?)(\s*)#\s*skip\s+if\s+(.*)$")


def doctest_skip_parser(func):
    """Decorator replaces custom skip test markup in doctests.

    Say a function has a docstring::

        >>> something # skip if not HAVE_AMODULE
        >>> something + else
        >>> something # skip if HAVE_BMODULE

    This decorator will evaluate the expression after ``skip if``.  If this
    evaluates to True, then the comment is replaced by ``# doctest: +SKIP``.
    If False, then the comment is just removed. The expression is evaluated in
    the ``globals`` scope of `func`.

    For example, if the module global ``HAVE_AMODULE`` is False, and module
    global ``HAVE_BMODULE`` is False, the returned function will have
    docstring::
        >>> something # doctest: +SKIP
        >>> something + else
        >>> something

    """
    lines = func.__doc__.split('\n')
    new_lines = []
    for line in lines:
        match = SKIP_RE.match(line)
        if match is None:
            new_lines.append(line)
            continue
        code, space, expr = match.groups()
        if eval(expr, func.__globals__):
            code = code + space + "# doctest: +SKIP"
        new_lines.append(code)
    func.__doc__ = "\n".join(new_lines)
    return func

###
# In some cases (e.g., on Travis), we want to use a virtual frame-buffer for
# testing. The following decorator runs the tests under xvfb (mediated by
# xvfbwrapper) conditioned on an environment variable (that we set in
# .travis.yml for these cases):
use_xvfb = os.environ.get('TEST_WITH_XVFB', False)
is_windows = platform.system().lower() == "windows"


def xvfb_it(my_test):
    """Run a test with xvfbwrapper."""
    # When we use verbose testing we want the name:
    fname = my_test.__name__

    def test_with_xvfb(*args, **kwargs):
        if use_xvfb:
            from xvfbwrapper import Xvfb
            display = Xvfb(width=1920, height=1080)
            display.start()
        my_test(*args, **kwargs)
        if use_xvfb:
            display.stop()
    # Plant it back in and return the new function:
    test_with_xvfb.__name__ = fname
    return test_with_xvfb if not is_windows else my_test


def set_random_number_generator(seed_v=1234):
    """Decorator to use a fixed value for the random generator seed.

    This will make the tests that use random functions reproducible.

    """
    def _set_random_number_generator(func):
        def _set_random_number_generator_wrapper(pytestconfig, *args, **kwargs):
            rng = np.random.default_rng(seed_v)
            kwargs['rng'] = rng
            signature = inspect.signature(func)
            if pytestconfig and 'pytestconfig' in signature.parameters.keys():
                output = func(pytestconfig, *args, **kwargs)
            else:
                output = func(*args, **kwargs)
            return output
        return _set_random_number_generator_wrapper
    return _set_random_number_generator



import pytest

import numpy as np
import numpy.testing as npt

from dipy.data import get_fnames
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package('tensorflow', min_version='2.0.0')

if have_tf:
    from dipy.nn.evac import EVACPlus


@pytest.mark.skipif(not have_tf, reason='Requires TensorFlow')
def test_default_weights():
    file_path = get_fnames('evac_test_data')
    input_arr = np.load(file_path)['input'][0]
    output_arr = np.load(file_path)['output'][0]

    evac_model = EVACPlus()
    results_arr = evac_model.predict(input_arr, np.eye(4), return_prob=True)
    npt.assert_almost_equal(results_arr, output_arr, decimal=4)


@pytest.mark.skipif(not have_tf, reason='Requires TensorFlow')
def test_default_weights_batch():
    file_path = get_fnames('evac_test_data')
    input_arr = np.load(file_path)['input']
    output_arr = np.load(file_path)['output']
    input_arr = [T1 for T1 in input_arr]

    evac_model = EVACPlus()
    fake_affine = np.array([np.eye(4), np.eye(4)])
    fake_voxsize = np.ones((2, 3))
    results_arr = evac_model.predict(input_arr, fake_affine, fake_voxsize, batch_size=2, return_prob=True)
    npt.assert_almost_equal(results_arr, output_arr, decimal=4)


@pytest.mark.skipif(not have_tf, reason='Requires TensorFlow')
def test_T1_error():
    T1 = np.ones((3, 32, 32, 32))
    evac_model = EVACPlus()
    fake_affine = np.array([np.eye(4), np.eye(4), np.eye(4)])
    npt.assert_raises(ValueError, evac_model.predict, T1, fake_affine)


import os

import numpy as np
import numpy.testing as npt
import pytest

from dipy.data import DATA_DIR
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.utils import create_nifti_header
from dipy.testing.decorators import use_xvfb
from dipy.tracking.streamline import Streamlines
from dipy.utils.optpkg import optional_package
from dipy.testing.decorators import set_random_number_generator

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import window

    from dipy.viz.horizon.app import horizon

skip_it = use_xvfb == 'skip'


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator()
def test_horizon_events(rng):
    # using here MNI template affine 2009a
    affine = np.array([[1., 0., 0., -98.],
                       [0., 1., 0., -134.],
                       [0., 0., 1., -72.],
                       [0., 0., 0., 1.]])

    data = 255 * rng.random((197, 233, 189))
    vox_size = (1., 1., 1.)

    images = [(data, affine)]
    # images = None
    from dipy.segment.tests.test_bundles import setup_module
    setup_module()
    from dipy.segment.tests.test_bundles import f1
    streamlines = f1.copy()
    streamlines._data += np.array([-98., -134., -72.])

    header = create_nifti_header(affine, data.shape, vox_size)
    sft = StatefulTractogram(streamlines, header, Space.RASMM)

    tractograms = [sft]

    # select all centroids and expand and click everything else
    # do not press the key shortcuts as vtk generates warning that
    # blocks recording
    fname = os.path.join(DATA_DIR, 'record_horizon.log.gz')

    horizon(tractograms=tractograms, images=images, pams=None,
            cluster=True, cluster_thr=5.0,
            random_colors=False, length_gt=0, length_lt=np.inf,
            clusters_gt=0, clusters_lt=np.inf,
            world_coords=True, interactive=True, out_png='tmp.png',
            recorded_events=fname)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator()
def test_horizon(rng):
    s1 = 10 * np.array([[0, 0, 0],
                        [1, 0, 0],
                        [2, 0, 0],
                        [3, 0, 0],
                        [4, 0, 0]], dtype='f8')

    s2 = 10 * np.array([[0, 0, 0],
                        [0, 1, 0],
                        [0, 2, 0],
                        [0, 3, 0],
                        [0, 4, 0]], dtype='f8')

    s3 = 10 * np.array([[0, 0, 0],
                        [1, 0.2, 0],
                        [2, 0.2, 0],
                        [3, 0.2, 0],
                        [4, 0.2, 0]], dtype='f8')

    streamlines = Streamlines()
    streamlines.append(s1)
    streamlines.append(s2)
    streamlines.append(s3)

    affine = np.array([[1., 0., 0., -98.],
                       [0., 1., 0., -134.],
                       [0., 0., 1., -72.],
                       [0., 0., 0., 1.]])

    data = 255 * rng.random((197, 233, 189))
    vox_size = (1., 1., 1.)

    streamlines._data += np.array([-98., -134., -72.])

    header = create_nifti_header(affine, data.shape, vox_size)
    sft = StatefulTractogram(streamlines, header, Space.RASMM)

    # only tractograms
    tractograms = [sft]
    images = None
    horizon(tractograms, images=images, cluster=True, cluster_thr=5,
            random_colors=False, length_lt=np.inf, length_gt=0,
            clusters_lt=np.inf, clusters_gt=0,
            world_coords=True, interactive=False)

    images = [(data, affine)]

    # tractograms in native coords (not supported for now)
    with npt.assert_raises(ValueError) as ve:
        horizon(tractograms, images=images, cluster=True, cluster_thr=5,
                random_colors=False, length_lt=np.inf, length_gt=0,
                clusters_lt=np.inf, clusters_gt=0,
                world_coords=False, interactive=False)

    msg = 'Currently native coordinates are not supported for streamlines.'
    npt.assert_(msg in str(ve.exception))

    # only images
    tractograms = None
    horizon(tractograms, images=images, cluster=True, cluster_thr=5,
            random_colors=False, length_lt=np.inf, length_gt=0,
            clusters_lt=np.inf, clusters_gt=0,
            world_coords=True, interactive=False)

    # no clustering tractograms and images
    horizon(tractograms, images=images, cluster=False, cluster_thr=5,
            random_colors=False, length_lt=np.inf, length_gt=0,
            clusters_lt=np.inf, clusters_gt=0,
            world_coords=True, interactive=False)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator(42)
def test_roi_images(rng):
    img1 = rng.random((5, 5, 5))
    img2 = np.zeros((5, 5, 5))
    img2[2, 2, 2] = 1
    img3 = np.zeros((5, 5, 5))
    img3[0, :, :] = 1
    images = [(img1, np.eye(4)), (img2, np.eye(4)), (img3, np.eye(4))]
    show_m = horizon(images=images, return_showm=True)
    analysis = window.analyze_scene(show_m.scene)
    npt.assert_equal(analysis.actors, 0)
    arr = window.snapshot(show_m.scene)
    report = window.analyze_snapshot(arr, colors=[(0, 0, 0), (255, 255, 255)])
    npt.assert_array_equal(report.colors_found, [True, True])
    show_m = horizon(images=images, roi_images=True, return_showm=True)
    analysis = window.analyze_scene(show_m.scene)
    npt.assert_equal(analysis.actors, 2)


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_small_horizon_import():
    from dipy.viz import horizon as Horizon
    if has_fury:
        assert Horizon == horizon
    else:
        npt.assert_raises(ImportError, Horizon)



import warnings

from functools import partial
import numpy as np

from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab import (HorizonTab, build_label, build_slider,
                                  build_switcher, build_checkbox)

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import colormap


class SlicesTab(HorizonTab):
    """Interaction tab for slice visualization.

    Attributes
    ----------
    name : str
        Name of the tab.
    """
    def __init__(self, slices_visualizer, slice_id=0,
                 force_render=lambda _element: None):
        super().__init__()
        self._visualizer = slices_visualizer

        if slice_id == 0:
            self._name = 'Image'
        else:
            self._name = f'Image {slice_id}'

        self._force_render = force_render
        self._tab_id = 0

        self.on_slice_change = lambda _tab_id, _x, _y, _z: None

        self._opacity_toggle = build_checkbox(
            labels=[''],
            checked_labels=[''],
            on_change=self._toggle_opacity)

        self._slice_opacity_label, self._slice_opacity = build_slider(
            initial_value=1.,
            max_value=1.,
            text_template='{ratio:.0%}',
            on_change=self._change_opacity,
            label='Opacity'
        )

        self._slice_x_label, self._slice_x = build_slider(
            initial_value=self._visualizer.selected_slices[0],
            max_value=self._visualizer.data_shape[0] - 1,
            text_template='{value:.0f}',
            label='X Slice'
        )

        self._change_slice_x = partial(self._change_slice_value,
                                       selected_slice=self._slice_x)
        self._adjust_slice_x = partial(self._change_slice_value,
                                       selected_slice=self._slice_x,
                                       sync_slice=True)
        self._visualize_slice_x = partial(
            self._visualizer.slice_actors[0].display_extent,
            y1=0, y2=self._visualizer.data_shape[1] - 1,
            z1=0, z2=self._visualizer.data_shape[2] - 1)

        self._slice_x.obj.on_value_changed = self._adjust_slice_x
        self._slice_x.obj.on_moving_slider = self._change_slice_x

        self._change_slice_visibility_x = partial(
            self._update_slice_visibility, selected_slice=self._slice_x,
            actor=self._visualizer.slice_actors[0])

        self._slice_x_toggle = build_checkbox(
            labels=[''],
            checked_labels=[''],
            on_change=self._change_slice_visibility_x)

        self._slice_y_label, self._slice_y = build_slider(
            initial_value=self._visualizer.selected_slices[1],
            max_value=self._visualizer.data_shape[1] - 1,
            text_template='{value:.0f}',
            label='Y Slice'
        )

        self._change_slice_y = partial(self._change_slice_value,
                                       selected_slice=self._slice_y)
        self._adjust_slice_y = partial(self._change_slice_value,
                                       selected_slice=self._slice_y,
                                       sync_slice=True)
        self._visualize_slice_y = partial(
            self._visualizer.slice_actors[1].display_extent,
            x1=0, x2=self._visualizer.data_shape[0] - 1,
            z1=0, z2=self._visualizer.data_shape[2] - 1)

        self._slice_y.obj.on_value_changed = self._adjust_slice_y
        self._slice_y.obj.on_moving_slider = self._change_slice_y

        self._change_slice_visibility_y = partial(
            self._update_slice_visibility, selected_slice=self._slice_y,
            actor=self._visualizer.slice_actors[1])

        self._slice_y_toggle = build_checkbox(
            labels=[''],
            checked_labels=[''],
            on_change=self._change_slice_visibility_y)

        self._slice_z_label, self._slice_z = build_slider(
            initial_value=self._visualizer.selected_slices[2],
            max_value=self._visualizer.data_shape[2] - 1,
            text_template='{value:.0f}',
            label='Z Slice'
        )

        self._change_slice_z = partial(self._change_slice_value,
                                       selected_slice=self._slice_z)
        self._adjust_slice_z = partial(self._change_slice_value,
                                       selected_slice=self._slice_z,
                                       sync_slice=True)
        self._visualize_slice_z = partial(
            self._visualizer.slice_actors[2].display_extent,
            x1=0, x2=self._visualizer.data_shape[0] - 1,
            y1=0, y2=self._visualizer.data_shape[1] - 1)

        self._slice_z.obj.on_value_changed = self._adjust_slice_z
        self._slice_z.obj.on_moving_slider = self._change_slice_z

        self._change_slice_visibility_z = partial(
            self._update_slice_visibility, selected_slice=self._slice_z,
            actor=self._visualizer.slice_actors[2])

        self._slice_z_toggle = build_checkbox(
            labels=[''],
            checked_labels=[''],
            on_change=self._change_slice_visibility_z)

        self._voxel_label = build_label(text='Voxel', is_horizon_label=True)
        self._voxel_data = build_label(text='', is_horizon_label=True)

        self._visualizer.register_picker_callback(self._change_picked_voxel)

        self.register_elements(
            self._opacity_toggle,
            self._slice_opacity_label,
            self._slice_opacity,
            self._slice_x_toggle,
            self._slice_x_label,
            self._slice_x,
            self._slice_y_toggle,
            self._slice_y_label,
            self._slice_y,
            self._slice_z_toggle,
            self._slice_z_label,
            self._slice_z,
            self._voxel_label,
            self._voxel_data
        )

        if not self._visualizer.rgb:
            self._intensities_label, self._intensities = build_slider(
                initial_value=self._visualizer.intensities_range,
                min_value=self._visualizer.volume_min,
                max_value=self._visualizer.volume_max,
                text_template='{value:.0f}',
                on_change=self._change_intensity,
                label='Intensities',
                is_double_slider=True
            )

            self._supported_colormap = [
                {'label': 'Gray', 'value': 'gray'},
                {'label': 'Bone', 'value': 'bone'},
                {'label': 'Cividis', 'value': 'cividis'},
                {'label': 'Inferno', 'value': 'inferno'},
                {'label': 'Magma', 'value': 'magma'},
                {'label': 'Viridis', 'value': 'viridis'},
                {'label': 'Jet', 'value': 'jet'},
                {'label': 'Pastel 1', 'value': 'Pastel1'},
                {'label': 'Distinct', 'value': 'dist'},

            ]

            self._colormap_switcher_label, self._colormap_switcher = \
                build_switcher(
                    items=self._supported_colormap,
                    label='Colormap',
                    on_value_changed=self._change_color_map
                )

            self.register_elements(
                self._intensities_label,
                self._intensities,
                self._colormap_switcher_label,
                self._colormap_switcher,
            )

            if len(self._visualizer.data_shape) == 4:
                self._volume_label, self._volume = build_slider(
                    initial_value=0,
                    max_value=self._visualizer.data_shape[-1] - 1,
                    on_moving_slider=self._change_volume,
                    text_template='{value:.0f}',
                    label='Volume'
                )
                self.register_elements(self._volume_label, self._volume)

    def _change_color_map(self, _idx, _value):
        self._update_colormap()
        self._force_render(self)

    def _change_intensity(self, slider):
        self._intensities.selected_value[0] = slider.left_disk_value
        self._intensities.selected_value[1] = slider.right_disk_value
        self._update_colormap()

    def _change_opacity(self, slider):
        self._slice_opacity.selected_value = slider.value
        self._update_opacities()

    def _toggle_opacity(self, checkbox):
        if '' in checkbox.checked_labels:
            self._slice_opacity.selected_value = 1
            self._slice_opacity.obj.value = 1
        else:
            self._slice_opacity.selected_value = 0
            self._slice_opacity.obj.value = 0
        self._update_opacities()

    def _change_picked_voxel(self, message):
        self._voxel_data.obj.message = message
        self._voxel_data.selected_value = message

    def _change_slice_value(
            self, slider, selected_slice, sync_slice=False):
        selected_slice.selected_value = int(np.rint(slider.value))
        if not sync_slice:
            self.on_slice_change(
                self._tab_id,
                self._slice_x.selected_value,
                self._slice_y.selected_value,
                self._slice_z.selected_value
            )
        self._visualize_slice_x(x1=self._slice_x.selected_value,
                                x2=self._slice_x.selected_value)
        self._visualize_slice_y(y1=self._slice_y.selected_value,
                                y2=self._slice_y.selected_value)
        self._visualize_slice_z(z1=self._slice_z.selected_value,
                                z2=self._slice_z.selected_value)

    def _update_slice_visibility(
            self, checkboxes, selected_slice, actor, visibility=None):

        if checkboxes is not None and '' in checkboxes.checked_labels:
            visibility = True
        elif visibility is None:
            visibility = False

        selected_slice.visibility = visibility
        selected_slice.obj.set_visibility(visibility)
        actor.SetVisibility(visibility)

    def _change_volume(self, slider):
        value = int(np.rint(slider.value))
        if value != self._volume.selected_value:
            visible_slices = (
                self._slice_x.selected_value, self._slice_y.selected_value,
                self._slice_z.selected_value)
            valid_vol = self._visualizer.change_volume(
                self._volume.selected_value, value,
                [self._intensities.selected_value[0],
                 self._intensities.selected_value[1]], visible_slices)
            if not valid_vol:
                warnings.warn(
                    f'Volume N°{value} does not have any contrast. Please, '
                    'check the value ranges of your data. Returning to '
                    'previous volume.')
                self._volume.obj.value = self._volume.selected_value
            else:
                intensities_range = self._visualizer.intensities_range

                # Updating the colormap
                self._intensities.selected_value[0] = intensities_range[0]
                self._intensities.selected_value[1] = intensities_range[1]
                self._update_colormap()

                # Updating intensities slider
                self._intensities.obj.initial_values = intensities_range
                self._intensities.obj.min_value = (
                    self._visualizer.volume_min)
                self._intensities.obj.max_value = (
                    self._visualizer.volume_max)
                self._intensities.obj.update(0)
                self._intensities.obj.update(1)

                # Updating opacities
                self._update_opacities()

                # Updating visibilities
                slices = [self._slice_x, self._slice_y, self._slice_z]
                for i, s in enumerate(slices):
                    self._update_slice_visibility(
                        None,
                        s,
                        self._visualizer.slice_actors[i],
                        s.visibility
                    )

                self._volume.selected_value = value
                self._force_render(self)

    def _update_colormap(self):
        if self._colormap_switcher.selected_value[1] == 'dist':
            rgb = colormap.distinguishable_colormap(nb_colors=256)
            rgb = np.asarray(rgb)
        else:
            rgb = colormap.create_colormap(
                np.linspace(self._intensities.selected_value[0],
                            self._intensities.selected_value[1], 256),
                name=self._colormap_switcher.selected_value[1], auto=True)
        num_lut = rgb.shape[0]

        lut = colormap.LookupTable()
        lut.SetNumberOfTableValues(num_lut)
        lut.SetRange(self._intensities.selected_value[0],
                     self._intensities.selected_value[1])
        for i in range(num_lut):
            r, g, b = rgb[i]
            lut.SetTableValue(i, r, g, b)
        lut.SetRampToLinear()
        lut.Build()

        for slice_actor in self._visualizer.slice_actors:
            slice_actor.output.SetLookupTable(lut)
            slice_actor.output.Update()

    def _update_opacities(self):
        for slice_actor in self._visualizer.slice_actors:
            slice_actor.GetProperty().SetOpacity(
                self._slice_opacity.selected_value)

    def on_tab_selected(self):
        self._slice_x.obj.set_visibility(self._slice_x.visibility)
        self._slice_y.obj.set_visibility(self._slice_y.visibility)
        self._slice_z.obj.set_visibility(self._slice_z.visibility)

    def update_slices(self, x_slice, y_slice, z_slice):
        """Updates slicer positions.

        Parameters
        ----------
        x_slice: float
            x-value where the slicer should be placed
        y_slice: float
            y-value where the slicer should be placed
        z_slice: float
            z-value where the slicer should be placed
        """
        if not self._slice_x.obj.value == x_slice:
            self._slice_x.obj.value = x_slice

        if not self._slice_y.obj.value == y_slice:
            self._slice_y.obj.value = y_slice

        if not self._slice_z.obj.value == z_slice:
            self._slice_z.obj.value = z_slice

    def build(self, tab_id, _tab_ui):
        """Build all the elements under the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        tab_ui : TabUI
            FURY TabUI object for tabs panel.

        Notes
        -----
        tab_ui will removed once every all tabs adapt new build architecture.
        """
        self._tab_id = tab_id

        x_pos = .02
        self._opacity_toggle.position = (x_pos, .85)
        self._slice_x_toggle.position = (x_pos, .62)
        self._slice_y_toggle.position = (x_pos, .38)
        self._slice_z_toggle.position = (x_pos, .15)

        x_pos = .05
        self._slice_opacity_label.position = (x_pos, .85)
        self._slice_x_label.position = (x_pos, .62)
        self._slice_y_label.position = (x_pos, .38)
        self._slice_z_label.position = (x_pos, .15)

        x_pos = .10
        self._slice_opacity.position = (x_pos, .85)
        self._slice_x.position = (x_pos, .62)
        self._slice_y.position = (x_pos, .38)
        self._slice_z.position = (x_pos, .15)

        x_pos = .52
        self._voxel_label.position = (x_pos, .38)

        x_pos = .60
        self._voxel_data.position = (x_pos, .38)

        if not self._visualizer.rgb:

            x_pos = .52
            self._intensities_label.position = (x_pos, .85)
            self._colormap_switcher_label.position = (x_pos, .56)

            x_pos = .60
            self._intensities.position = (x_pos, .85)
            self._colormap_switcher.position = [
                (x_pos, .54), (0.63, .54), (0.69, .54)
            ]

            if len(self._visualizer.data_shape) == 4:
                x_pos = .52
                self._volume_label.position = (x_pos, .15)

                x_pos = .60
                self._volume.position = (x_pos, .15)

    @property
    def name(self):
        """Name of the tab.
        """
        return self._name

    @property
    def tab_id(self):
        """Id of the tab.
        """
        return self._tab_id

    @property
    def actors(self):
        """visualization actors controlled by tab.
        """
        return self._visualizer.slice_actors
